"""
 Session Description - RFC 4566
 Very basic SDP parser
 ------------------------------------------------------------------------------------------------------
 type_    Dictionary key          Format of the value
 ======  ======================  ======================================================================
 v       "protocol_version"      version_number
 o       "origin"                ("user", session_id, session_version, "net_type", "addr_type", "addr")
 s       "sessionname"           "session name"
 t & r   "time"                  (starttime, stoptime, [repeat,repeat, ...])
                                    where repeat = (interval,duration,[offset,offset, ...])
 a       "attribute"             "value of attribute"
 b       "bandwidth"             (mode, bitspersecond)
 i       "information"           "value"
 e       "email"                 "email-address"
 u       "URI"                   "uri"
 p       "phone"                 "phone-number"
 c       "connection"            ("net_type", "addr_type", "addr", ttl, groupsize)
 z       "timezone adjustments"  [(adj-time,offset), (adj-time,offset), ...]
 k       "encryption"            ("method","value")
 m       "media"                 [media-description, media-description, ... ]
                                     see next table for media description structure
 ======  ======================  ======================================================================
"""
import re
from typing import Optional


class SDP(dict):
    """
    SDP Parser class.
    Takes an sdp content as an input and split it into various sections,
    including the different available medias.
    """

    def __init__(self, data: str):
        super().__init__()
        self.current_media = None

        self.parsers = {
            'v': self._parse_version,
            'o': self._parse_origin,
            's': self._parse_session_name,
            'i': self._parse_info,
            'u': self._parse_uri,
            'e': self._parse_email,
            'p': self._parse_phone,
            'c': self._parse_connection,
            'k': self._parse_encryption,
            'z': self._parse_timezone,
            'm': self._parse_media,
            'b': self._parse_bandwidth,
            't': self._parse_time,
            'r': self._parse_repeats,
            'a': self._parse_attributes,
        }

        for line in data.splitlines():
            self._parseline(line)

    def _parse_version(self, value):
        value = int(value)
        # Only version 0 allowed
        assert value == 0, 'only SDP version 0 supported'
        self['version'] = int(value)

    def _parse_origin(self, value):
        self['origin'] = value

    def _parse_session_name(self, value):
        self['sessionName'] = value

    def _parse_info(self, value):
        self['information'] = value

    def _parse_uri(self, value):
        self['URI'] = value

    def _parse_email(self, value):
        self['email'] = value

    def _parse_phone(self, value):
        self['phone'] = value

    def _parse_connection(self, value):
        self._element['connection'] = value

    def _parse_encryption(self, value):
        method, value = re.match(r"^(clear|base64|uri|prompt)(?:[:](.*))?$", value).groups()
        self._element["encryption"] = (method, value)

    def _parse_timezone(self, value):
        adjustments = []
        while value.strip() != "":
            adjtime, offset, offsetunit, value = re.match(r"^ *(\d+) +([+-]?\d+)([dhms])? *?(.*)$", value).groups()
            adjtime = int(adjtime)
            offset = int(offset) * {None: 1, "s": 1, "m": 60, "h": 3600, "d": 86400}[offsetunit]
            adjustments.append((adjtime, offset))

        self._element['timezoneAdjustments'] = adjustments

    def _parse_media(self, value):
        media, port, numports, protocol, fmt = re.match(
            r"^(audio|video|text|application|message) +(\d+)(?:[/](\d+))? +([^ ]+) +(.+)$", value).groups()

        port = int(port)

        if numports is None:
            numports = 1
        else:
            numports = int(numports)

        self.current_media = {
            'type': media,
            'port': port,
            'numPorts': numports,
            'protocol': protocol,
            'format': fmt
        }
        self.setdefault('medias', []).append(self.current_media)

    def _parse_bandwidth(self, value):
        mode, rate = \
            re.match(r"^ *((?:AS)|(?:CT)|(?:X-[^:]+)):(\d+) *$", value).groups()
        bitspersecond = int(rate) * 1000

        self._element['bandwidth'] = (mode, bitspersecond)

    def _parse_time(self, value):
        start, stop = [int(x) for x in re.match(r"^ *(\d+) +(\d+) *$", value).groups()]
        self._element['time'] = (start, stop)

    def _parse_repeats(self, value):
        terms = re.split(r"\s+", value)
        parsedterms = []
        for term in terms:
            value, unit = re.match(r"^\d+([dhms])?$", term).groups()
            value = int(value) * {None: 1, "s": 1, "m": 60, "h": 3600, "d": 86400}[unit]
            parsedterms.append(value)

        interval, duration = parsedterms[0], parsedterms[1]
        offsets = parsedterms[2:]
        self._element['repeats'] = (interval, duration, offsets)

    def _parse_attributes(self, value):
        # Attributes are a=<attrname>:<specific content>
        attr, content = value.split(':', 1) if ':' in value else (value, None)
        attributes = self._element.setdefault('attributes', {})

        # # Special cases
        if attr == 'framerate':
            # Content is a framerate as float
            content = float(content)
        elif attr == 'framesize':
            pt, width, height = re.match(r"^ *(\d+) *(\d+)-(\d+) *$", content).groups()
            content = {
                'pt': int(pt),
                'width': int(width),
                'height': int(height)
            }
        elif attr == 'rtpmap':
            pt, enc, clock = re.match(r"^ *(\d+) *(\S+)/(\d+) *$", content).groups()
            content = {
                'pt': int(pt),
                'encoding': enc,
                'clockRate': int(clock)
            }
        elif attr == 'fmtp':
            pt, opts = content.split(None, 1)
            content = {
                'pt': int(pt)
            }
            for opt in opts.split(';'):
                if not opt.strip():
                    # Empty, probably a wrong semicolon at the end...
                    continue
                k, v = opt.split('=', 1)
                content[k.strip()] = v.strip()

        attributes[attr] = content

    def _parse_unknown(self, type_, value):
        self._element.setdefault('unknowns', []).append((type_, value))

    @property
    def _element(self):
        return self if self.current_media is None else self.current_media

    def _parseline(self, line):
        match = re.match("^(.)=(.*)", line)

        if match:
            type_, value = match.group(1), match.group(2)

            if type_ in self.parsers:
                try:
                    self.parsers[type_](value)
                except Exception:  # pylint: disable=broad-except
                    self._parse_unknown(type_, value)
            else:
                self._parse_unknown(type_, value)

    @staticmethod
    def mix_url_control(base: str, ctrl) -> str:
        """
        Given a base URL and a control attribute,
        build an URL to be used during SETUP.
        :param base: Base URL (either from user or returned in Content-Base)
        :param ctrl: Control attributes for given media (or global)
        :return: URL
        """
        if not ctrl or ctrl == '*':
            return base

        if ctrl.startswith('rtsp://'):
            return ctrl

        if not ctrl.startswith('/') and not base.endswith('/'):
            return base + '/' + ctrl

        return base + ctrl

    def get_media(self, media_type='video', media_idx=0):
        """
        Return the Nth media description matching requested type
        :param media_type:
        :param media_idx:
        :return:
        """
        current_idx = 0

        for media in self.get('medias', []):
            if media['type'] != media_type:
                continue

            if current_idx < media_idx:
                current_idx += 1
                continue

            # Found it!
            return media

        return None

    def setup_url(self, base_url: str, media_type='video', media_idx=0) -> str:
        """
        Return the URL to be used for setup.
        :param base_url: (url requested or returned base url)
        :param media_type: audio|video|text|application|message
        :param media_idx: index if multiple medias are available
        :return: corrected URL
        """
        # Check global control
        base_url = self.mix_url_control(base_url, self.get('attributes', {}).get('control'))

        # Look for media
        media = self.get_media(media_type, media_idx)
        if media:
            return self.mix_url_control(base_url, media.get('attributes', {}).get('control'))

        # Not found in medias
        return base_url

    def media_clock_rate(self, media_type='video', media_idx=0) -> Optional[int]:
        """
        Return clock rate of given media
        """
        media = self.get_media(media_type, media_idx)
        if media:
            return media.get('attributes', {}).get('rtpmap', {}).get('clockRate')
        return None

    def media_payload_type(self, media_type='video', media_idx=0) -> Optional[int]:
        """
        Return clock rate of given media
        """
        media = self.get_media(media_type, media_idx)
        if media:
            return media.get('attributes', {}).get('rtpmap', {}).get('pt')
        return None

    def guess_h264_props(self, media_idx=0):
        """
        Try to guess H264 `sprop-parameter-sets`
        :param media_idx:
        :return: props string
        """
        media = self.get_media(media_type='video', media_idx=media_idx)
        if media:
            return media.get('attributes', {}).get('fmtp', {}).get('sprop-parameter-sets')
        return None



"""
RTP Transport module.
---------------------

In this file lies the main part of:
 - handling RTP/RTCP data
 - sending RTCP reports
 - allowing clients to subscribe to RTP/RTCP packets

Then in different files can be found different transport implementation.
"""
from typing import Type

from .base import RTPTransport, RTPTransportClient  # noqa
from .tcp import TCPTransport
from .udp import UDPTransport


def transport_for_scheme(scheme: str) -> Type[RTPTransport]:
    """Return transport type based on scheme"""
    transport_class = {
        'rtsp': UDPTransport,
        'rtspt': TCPTransport
    }.get(scheme)

    if not transport_class:
        raise ValueError(f'invalid URL scheme `{scheme}`')

    return transport_class



"""
RTCP Statistics module
"""

import logging
import random
from math import fabs
from time import time
from typing import Optional

from dpkt.rtp import RTP

from .parser import RTCP, SRReport, ts_to_ntp, RR, SDES, BYE, SR

_logger = logging.getLogger('rtcp.sink')

RTP_SEQ_MOD = (1 << 16)
MAX_MISORDER = 100
MAX_DROPOUT = 3000
UINT32_MASK = 0xFFFFFFFF
UINT32_MASK = 0xFFFFFFFF
NTP_MASK = 0xFFFF


class RTCPStats:
    """
    RTCP Statistics and response builder
    ------------------------------------

    This class builds statistics upon reception of RTP frames,
    handles reception of remote RTCP report from the server,
    and crafts RTCP report packet.

    This is a simplified version not handling multicast yet.
    """
    def __init__(self, name='unknown'):
        self.name = name
        self.ssrc = random.randint(0, 2**32)
        self.ntp = self.ntp_base = time()

        self.received = self.expected_prior = self.received_prior = self.transit = self.jitter = 0
        self.lost = self.fraction = self.pkt_count = self.oct_count = self.timeout = 0
        self.probation = self.last_received = self.ts = self.ts_base = 0

        self.maxseq = self.bad_seq = self.base_seq = self.cycles = None
        self.last_ts = self.last_ntp = self.rtcp_delay = None

    @property
    def cname(self) -> bytes:
        """
        Return CNAME for RR reporting
        """
        return f'{self.ssrc}@{self.name}'.encode()

    @property
    def extended_seq(self) -> int:
        """Extended sequence number, taking cycles into account"""
        return self.cycles + self.maxseq

    @property
    def lsr(self) -> int:
        """Build LSR report value"""
        ntp1, ntp2 = ts_to_ntp(self.last_ntp)
        return ((ntp1 & NTP_MASK) << 16) | ((ntp2 >> 16) & NTP_MASK)

    @property
    def dlsr(self) -> int:
        """Build DLSR report value"""
        return max(0, min(UINT32_MASK, int((time() - self.last_ntp) * 65536)))

    def init_seq(self, seq):
        """
        Initialize the seq using the newly received seq of RTP packet.
        """
        self.base_seq = self.maxseq = seq
        self.bad_seq = seq - 1
        self.cycles = self.received = self.received_prior = self.expected_prior = 0

    def update_seq(self, seq):
        """
        Update the source properties based on received RTP packet's seq.
        """
        seq_delta = seq - self.maxseq
        if self.probation > 0:
            if seq == self.maxseq + 1:
                self.probation = self.probation - 1
                self.maxseq = seq
                if self.probation == 0:
                    # reset
                    self.init_seq(seq)
                    self.received = self.received + 1
                    return
            else:
                # next packet should be in sequence
                self.probation = 1
                self.maxseq = seq
            return

        if seq_delta < MAX_DROPOUT:
            # in order, within reasonable gap
            if seq < self.maxseq:
                self.cycles += RTP_SEQ_MOD
            self.maxseq = seq
        elif seq_delta <= RTP_SEQ_MOD - MAX_MISORDER:
            # Huge gap
            if seq == self.bad_seq:
                # Sequence was either reset or changed
                self.init_seq(seq)
            else:
                self.bad_seq = (seq + 1) & (RTP_SEQ_MOD - 1)
                return

        # Count this packet
        self.received += 1

    def update_jitter(self, ts):
        """
        Update the jitter based on ts and arrival (in ts units).
        """
        transit = int(self.ts_now - ts)
        d, self.transit = int(fabs(transit - self.transit)), transit
        self.jitter += (1 / 16) * (d - self.jitter)

    def update_lost_expected(self):
        """
        Update the number of packets expected and lost for reporting.
        """
        expected = self.extended_seq - self.base_seq + 1
        received = self.received

        received_interval = received - self.received_prior
        expected_interval = expected - self.expected_prior
        lost_interval = expected_interval - received_interval

        self.expected_prior = expected
        self.received_prior = received
        self.lost = expected - received

        if expected_interval == 0 or lost_interval <= 0:
            self.fraction = 0
        else:
            self.fraction = (lost_interval << 8) // expected_interval

    @staticmethod
    def rtcp_interval(initial=False) -> float:
        """
        Simplified RTCP interval calculator.
        TODO implement multicast. In unicast, it does not really matter...
        :param initial: Is this the first interval (shorter)
        """
        return (2.5 if initial else 5.0) * (random.random() + 0.5) / 1.21828

    @property
    def ts_now(self):
        """The current RTP timestamp in ts unit based on current time."""
        ts = self.ts
        if self.ntp != self.ntp_base:
            ts += (time() - self.ntp) * ((self.ts - self.ts_base) / (self.ntp - self.ntp_base))
        return int(ts) & UINT32_MASK

    def update(self, pkt: RTP):
        """
        Called externally to handle a received RTP packet and update statistics.
        """
        seq = pkt.seq
        if self.base_seq is None:
            self.init_seq(seq)
        self.update_seq(seq)
        self.update_jitter(pkt.ts)
        self.last_received = time()

    def handle_rtcp(self, rtcp: RTCP):
        """
        RTCP packet received
        """
        for p in rtcp.packets:
            if isinstance(p, SR):
                self.last_ts = p.ts
                self.last_ntp = p.ntp

    def build_rtcp(self, send_bye=False) -> Optional[RTCP]:
        """
        Build an RTCP packet based on current situation.

        There may be not enough info yet to return anything!

        :param send_bye: Add a bye header
        """
        if self.last_ntp is None:
            # Not received a first SR report yet...
            return None

        if self.base_seq is None:
            # Not received an RTP packet yet...
            return None

        self.update_lost_expected()

        rtcp = RTCP()

        # Add receiver report
        rr = RR(self.ssrc, reports=[
            SRReport(
                ssrc=self.ssrc, flost=self.fraction, clost=self.lost,
                hseq=self.extended_seq, jitter=int(self.jitter),
                lsr=self.lsr, dlsr=self.dlsr)
        ])
        rtcp.packets.append(rr)

        # Add SDES identity
        sdes = SDES(items=[(1, [(SDES.CNAME, self.cname)])])
        rtcp.packets.append(sdes)

        # Add bye if requested
        if send_bye:
            bye = BYE(ssrcs=[self.ssrc])
            rtcp.packets.append(bye)

        return rtcp



"""
RTSP Module errors
"""


class RTSPError(Exception):
    """Base RTSP error"""
    type = 'unknown'

    def reason(self) -> dict:
        """
        Structured data for RTSP errors, including type,
        and potentially type-specific extra values.
        """
        return {
            'type': self.type
        }


class RTSPResponseError(RTSPError):
    """
    Error upon response. If the response is provided, will be printed
    """
    type = 'response'

    def __init__(self, msg, response=None):
        self.response = response
        super().__init__(msg, self.reason())

    def __str__(self):
        msg = f'REASON: {super().__str__()}'
        if self.response:
            msg += f'\nRESPONSE: {self.response}'
            if self.response.content:
                msg += f'\nCONTENT: {self.response.content}'
        return msg

    def reason(self):
        r = super().reason()
        if self.response:
            r['status'] = self.response.status
            r['message'] = self.response.status_msg
        return r


class RTSPConnectionError(RTSPError):
    """
    RTSP connection error
    """
    type = 'connection'


class RTSPTimeoutError(RTSPError):
    """
    A request timed out (no answer we could understand)
    """
    type = 'timeout'



"""
RTCP Parsing module
-------------------
This should really just be a lib call like we do for RTP (using dpkt),
but there is no light/easily importable RTCP parser available...
"""
from abc import ABC, abstractmethod
from struct import unpack, pack
from typing import Tuple, Dict, Type, Optional, List

TS_OFFSET_1900 = 2208988800


def ts_to_ntp(ts: float) -> Tuple[int, int]:
    """
    Convert from time.time() output to NTP (seconds, fraction).
    """
    ts += TS_OFFSET_1900  # From 1900 to 1970
    seconds, fraction = divmod(ts, 1)
    return int(seconds), int(fraction * 2**32)


def ntp_to_ts(seconds: int, fraction: int) -> float:
    """
    Convert from NTP (seconds, fraction) to time similar to time.time() output.
    """
    return (seconds + fraction / 2**32) - TS_OFFSET_1900


RTCP_TYPES: Dict[int, Type['RTCPPacket']] = {}


class RTCPPacket:
    """
    Sub-content of a generic RTCP container
    """
    pt = 0

    def __init_subclass__(cls, **kwargs):
        """
        Register RTCP type in the RTCP_TYPES registry
        """
        super().__init_subclass__(**kwargs)

        if cls.pt:
            RTCP_TYPES[cls.pt] = cls

    @classmethod
    @abstractmethod
    def unpack(cls, px, pt, p_len, data) -> 'RTCPPacket':
        """
        Unpack an RTCP sub-packet
        """

    @abstractmethod
    def __bytes__(self):
        """Serialize to bytes"""

    def pack(self, count, value, pad=False) -> bytes:
        """
        Turn an RTCP packet back into bytes
        """
        length = len(value) // 4 + (1 if len(value) % 4 != 0 else 0)
        px = 0x80 | (pad and len(value) % 4 != 0 and 0x20 or 0x00) | (count & 0x1f)
        header = pack('!BBH', px, self.pt, length)

        if not pad or len(value) % 4 == 0:
            padding = b''
        else:
            padding = b'\x00' * (4 - len(value) % 4 - 1) + pack('!B', 4 - len(value) % 4)
        return header + value + padding


class SRReport:
    """
    Sender/Receiver report content
    """

    def __init__(self, ssrc, flost, clost, hseq, jitter, lsr, dlsr):
        self.ssrc = ssrc
        self.flost = flost
        self.clost = clost
        self.hseq = hseq
        self.jitter = jitter
        self.lsr = lsr
        self.dlsr = dlsr

    @classmethod
    def unpack(cls, data: bytes) -> 'SRReport':
        """
        Transform a report into bytes
        """
        assert len(data) == 24
        ssrc, lost, hseq, jitter, lsr, dlsr = unpack('!IIIIII', data)
        flost, clost = (lost >> 24) & 0xFF, (lost & 0x00FFFFFF)
        return cls(ssrc, flost, clost, hseq, jitter, lsr, dlsr)

    def __repr__(self):
        return f'<SRReport ssrc={self.ssrc} seq={self.hseq} flost={self.flost}/256 cumul={self.clost} jitter={self.jitter}>'

    def __bytes__(self):
        return pack(
            '!IIIIII',
            self.ssrc,
            (self.flost << 24) | (self.clost & 0x00FFFFFF),
            self.hseq,
            self.jitter,
            self.lsr,
            self.dlsr
        )


class SRRR(RTCPPacket, ABC):
    """
    Base SR/RR class (most things in common)
    """

    def __init__(self, ssrc, extn=None, reports=None):
        self.ssrc = ssrc
        self.extn = extn
        self.reports = reports or []

    @classmethod
    def unpack(cls, px, pt, p_len, data):
        raise NotImplementedError

    @classmethod
    def parse_reports(cls, px, data):
        """
        Parse internal reports
        """
        reports = []
        for _ in range(px & 0x1F):
            assert len(data) >= 24
            report_data = data[:24]
            reports.append(SRReport.unpack(report_data))
            data = data[24:]
        return reports


class SR(SRRR):
    """
    SR sub-type
    """
    pt = 200

    def __init__(self, ssrc, ntp, ts=0, pkt_count=0, byte_count=0, extn=None, reports=None):
        super().__init__(ssrc=ssrc, extn=extn, reports=reports)
        self.ntp = ntp
        self.ts = ts
        self.pkt_count = pkt_count
        self.byte_count = byte_count

    @classmethod
    def unpack(cls, px, pt, p_len, data):
        ssrc, ntp1, ntp2, ts, pkt_count, byte_count = unpack('!IIIIII', data[:24])
        ntp = ntp_to_ts(ntp1, ntp2)
        reports = cls.parse_reports(px, data[24:])
        return cls(ssrc=ssrc, ntp=ntp, ts=ts, pkt_count=pkt_count, byte_count=byte_count, reports=reports)

    def __repr__(self):
        return f'<SR ssrc={self.ssrc} ts={self.ts} pkt={self.pkt_count} ntp={self.ntp} reports={self.reports}>'

    def __bytes__(self):
        ntp1, ntp2 = ts_to_ntp(self.ntp)
        value = pack('!IIIIII', self.ssrc, ntp1, ntp2, self.ts, self.pkt_count, self.byte_count)
        count = len(self.reports)
        for report in self.reports:
            value += bytes(report)
        if self.extn:
            value += self.extn

        return self.pack(count, value)


class RR(SRRR):
    """
    RR sub-type
    """
    pt = 201

    @classmethod
    def unpack(cls, px, pt, p_len, data):
        ssrc, = unpack('!I', data[:4])
        reports = cls.parse_reports(px, data[4:])
        return cls(ssrc, reports=reports)

    def __repr__(self):
        return f'<RR ssrc={self.ssrc} reports={self.reports}>'

    def __bytes__(self):
        value = pack('!I', self.ssrc)
        count = len(self.reports)

        for report in self.reports:
            value += bytes(report)

        if self.extn:
            value += self.extn

        return self.pack(count, value)


class SDES(RTCPPacket):
    """
    SDES sub-type
    """
    pt = 202

    CNAME, NAME, EMAIL, PHONE, LOC, TOOL, NOTE, PRIV = range(1, 9)

    def __init__(self, items=None):
        self.items = items or []

    @classmethod
    def unpack(cls, px, pt, p_len, data):
        items = []
        for _ in range(0, px & 0x1F):
            ssrc, = unpack('!I', data[:4])
            local_items = []
            data, count = data[4:], 0

            while len(data) >= 2:
                itype, ilen = unpack('!BB', data[:2])
                count += (2 + ilen)
                ivalue, data = data[2:2+ilen], data[2+ilen:]
                if itype == 0:
                    break
                local_items.append((itype, ivalue))
            if count % 4 != 0:
                data = data[(4-count % 4):]
            items.append((ssrc, local_items))

        return cls(items=items)

    def __repr__(self):
        return f'<SDES items={self.items}>'

    def __bytes__(self):
        value = b''
        count = len(self.items)

        for ssrc, items in self.items:
            chunk = pack('!I', ssrc)
            for n, v in items:
                chunk += pack('!BB', n, len(v) > 255 and 255 or len(v)) + v[:256]
            chunk += pack('!BB', 0, 0)  # to indicate end of items.
            if len(chunk) % 4 != 0:
                chunk += b'\x00' * (4 - len(chunk) % 4)
            value += chunk

        return self.pack(count, value)


class BYE(RTCPPacket):
    """
    BYE sub-type
    """
    pt = 203

    def __init__(self, ssrcs=None, reason=None):
        self.ssrcs = ssrcs or []
        self.reason = reason

    @classmethod
    def unpack(cls, px, pt, p_len, data):
        ssrcs, reason = [], None

        for _ in range(0, px & 0x01F):
            ssrc, = unpack('!I', data[:4])
            ssrcs.append(ssrc)
            data = data[4:]

        if data:
            rlen, = unpack('!B', data[:1])
            reason = data[1:1+rlen]

        return cls(ssrcs=ssrcs, reason=reason)

    def __repr__(self):
        return f'<BYE ssrcs={self.ssrcs} reason={self.reason}>'

    def __bytes__(self):
        value = b''
        count = len(self.ssrcs)
        for ssrc in self.ssrcs:
            value += pack('!I', ssrc)

        if self.reason:
            reason_len = min(len(self.reason), 255)
            value += pack('!B', reason_len) + self.reason[:reason_len]

        return self.pack(count, value)


class APP(RTCPPacket):
    """
    APP sub-type
    """
    pt = 204

    def __init__(self, subtype, ssrc, name, data=None):
        self.subtype = subtype
        self.ssrc = ssrc
        self.name = name
        self.data = data

    @classmethod
    def unpack(cls, px, pt, p_len, data):
        subtype = px & 0x1F
        ssrc, name = unpack('!I4s', data[:8])
        data = data[8:]

        return cls(subtype=subtype, ssrc=ssrc, name=name, data=data)

    def __repr__(self):
        return f'<APP stype={self.subtype} ssrc={self.ssrc} name={self.name} data_len={len(self.data)}>'

    def __bytes__(self):
        return self.pack(self.subtype, self.data)


class RTCP:
    """
    RTP Control Protocol Packet
    """

    def __init__(self, packets=None):
        self.packets: List[RTCPPacket] = packets or []

    @classmethod
    def unpack(cls, data: bytes) -> 'RTCP':
        """
        Parse a buffer into an RTCP instance
        """
        packets = []

        # Iterate on every packet found
        while data:

            # Unpack the header before deciding type
            px, pt, p_len = unpack('!BBH', data[:4])

            # Check version
            if px & 0xC0 != 0x80:
                raise ValueError('RTP version must be 2')

            # Check type exists
            if pt not in RTCP_TYPES:
                raise ValueError(f'Not an RTCP packet type: {pt}')

            # Split current bytes and bytes from next packet
            payload_length = 4 + p_len * 4
            if payload_length > len(data):
                raise ValueError(f'RTCP Packet truncated ({payload_length} > {len(data)})')

            payload, data = data[4:payload_length], data[payload_length:]

            if px & 0x20:
                # There is some padding to be removed
                payload = payload[:len(payload)-payload[-1]]

            packets.append(RTCP_TYPES[pt].unpack(px, pt, p_len, payload))

        return cls(packets=packets)

    def get(self, pt: int) -> Optional[RTCPPacket]:
        """
        Return first occurrence of given payload type, if any.
        """
        match = [p for p in self.packets if p.pt == pt]
        return None if not match else match[0]

    def __repr__(self):
        return f'RTCP({self.packets})'

    def __bytes__(self):
        result = b''
        for packet in self.packets:
            result += bytes(packet)
        return result



"""
UDP Socket transport
"""
import asyncio
import logging
import socket
from random import randint
from typing import Tuple

from aiortsp.rtcp.parser import RTCP
from aiortsp.rtsp.errors import RTSPError
from .base import RTPTransport

_logger = logging.getLogger('rtp.session')

DEFAULT_BUFFER_SIZE = 4 * 1024 * 1024


class DatagramSink(asyncio.DatagramProtocol):
    """
    Default sink implementation.
    """
    def __init__(self, receiver: 'UDPTransport'):
        self.receiver = receiver
        self.transport = None

    @property
    def local_port(self) -> int:
        """
        local port or identifier
        """
        if self.transport:
            return self.transport.get_extra_info('socket').getsockname()[1]
        return 0

    @property
    def running(self) -> bool:
        """
        A sink is running until marked as done
        """
        return self.transport is not None

    def sendto(self, data, remote):
        """
        Convenient wrapper
        """
        if self.transport:
            self.transport.sendto(data, remote)

    def close(self):
        """
        Convenient wrapper
        """
        if self.transport:
            self.transport.close()

    def connection_made(self, transport):
        self.transport = transport

        sock = self.transport.get_extra_info('socket')

        try:
            self.receiver.logger.debug(
                'setting socket send buffer size to: %s', self.receiver.send_buffer)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, self.receiver.send_buffer)
        except OSError as ex:
            # What can we do here?
            #  - On Linux it would just set the max.
            #  - On mac, if too big it will just fail
            self.receiver.logger.warning('could not set socket SO_SNDBUF: %s', ex)

        try:
            self.receiver.logger.debug(
                'setting socket receive buffer size to: %s', self.receiver.receive_buffer)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, self.receiver.receive_buffer)
        except OSError as ex:
            # What can we do here?
            #  - On Linux it would just set the max.
            #  - On mac, if too big it will just fail
            self.receiver.logger.warning('could not set socket SO_RCVBUF: %s', ex)

    def connection_lost(self, exc):
        self.transport = None


class RTCPSink(DatagramSink):
    """
    RTCP Report sink
    """
    def datagram_received(self, data, addr):
        self.receiver.handle_rtcp(data, addr)


class RTPSink(DatagramSink):
    """
    RTP sink
    """
    def datagram_received(self, data, addr):
        self.receiver.handle_rtp_data(data)


class UDPTransport(RTPTransport):
    """
    UDP Transport.
    --------------

    Uses a pair of consecutive UDP socket to receive RTP and RTCP.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.receive_buffer = kwargs.get('receive_buffer', DEFAULT_BUFFER_SIZE)
        self.send_buffer = kwargs.get('send_buffer', DEFAULT_BUFFER_SIZE)

        self.rtp_sink: RTPSink = None
        self.rtcp_sink: RTCPSink = None
        self.server_rtp = None
        self.server_rtcp = None
        self.rtcp_sender = None

    @classmethod
    def get_socket_pair(cls, bind_address=None, retry=10) -> Tuple[socket.socket, socket.socket]:
        """
        Try to allocate a pair of consecutive UDP ports.
        In theory, RTP and RTCP could have completely unrelated port numbers,
        but many server assume that RTCP = RTP + 1 (and even that RTP is always even...).
        :param bind_address: IP address to bind UDP ports. Default is ANY
        :param retry: number of attempts to perform before giving up.
        :return:
        """
        while retry > 0:
            retry -= 1
            rtp_sock = rtcp_sock = None

            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sock.bind((bind_address or '0.0.0.0', 0))

                rtp_port = sock.getsockname()[1]

                sock2 = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sock2.bind((bind_address or '0.0.0.0', rtp_port + 1 if (rtp_port % 2) == 0 else rtp_port - 1))

                # Done!
                if (rtp_port % 2) == 0:
                    return sock, sock2

                return sock2, sock

            except IOError:
                _logger.error('unable to create requested socket pair')
                if rtp_sock:
                    rtp_sock.close()
                if rtcp_sock:
                    rtcp_sock.close()

        raise IOError('unable to allocate 2 consecutive ports')

    async def prepare(self):
        rtp_transport = rtcp_transport = None
        try:
            loop = asyncio.get_event_loop()

            rtp_sock, rtcp_sock = await asyncio.wait_for(
                loop.run_in_executor(None, self.get_socket_pair), 10)

            # Try to create RTP endpoint
            rtp_transport, self.rtp_sink = await loop.create_datagram_endpoint(
                lambda: RTPSink(self),
                sock=rtp_sock
            )

            # Try to create RTCP endpoint
            rtcp_transport, self.rtcp_sink = await loop.create_datagram_endpoint(
                lambda: RTCPSink(self),
                sock=rtcp_sock
            )

            self.logger.info(
                'UDP Transport ready, will use ports %s-%s',
                self.rtp_sink.local_port,
                self.rtcp_sink.local_port,
            )

        except Exception:
            if rtp_transport:
                rtp_transport.close()

            if rtcp_transport:
                rtcp_transport.close()

            raise

    @property
    def running(self) -> bool:
        """
        True if both RTP and RTCP sinks are running.
        """
        return self.rtp_sink.running and self.rtcp_sink.running

    def handle_rtcp(self, data, sender):
        """
        Special case for RTCP: we may not have received any report yet
        :param data:
        :param sender:
        """
        if self.rtcp_sender is None:
            self.logger.info('received RTCP from %s', sender)
            self.rtcp_sender = sender
        self.handle_rtcp_data(data)

    def on_transport_request(self, headers: dict):
        rtp_port = self.rtp_sink.local_port
        rtcp_port = self.rtcp_sink.local_port
        headers['Transport'] = f'RTP/AVP;unicast;client_port={rtp_port}-{rtcp_port}'

    def on_transport_response(self, headers: dict):
        if 'transport' not in headers:
            raise RTSPError('error on SETUP: Transport not found')

        # Get server port
        fields = self.parse_transport_fields(headers['transport'])

        if 'server_port' in fields:
            self.server_rtp, self.server_rtcp = fields['server_port'].split('-', 1)
            self.server_rtp = int(self.server_rtp)
            self.server_rtcp = int(self.server_rtcp)
            self.logger.info('server RTP/RTCP ports: %s-%s', self.server_rtp, self.server_rtcp)

    def close(self, error=None):
        """
        Perform cleanup, which by default is closing both sinks.
        """
        super().close(error)
        self.rtp_sink.close()
        self.rtcp_sink.close()

    @staticmethod
    def send_upstream(sink, address, port, count=5, length=200):
        """
        Send dummy packets to RTP source port to force NAT traversal
        :param sink: socket where to send dummy traffic
        :param address: server address
        :param port: server port
        :param count: number of packets to be sent
        :param length: length of payload
        """
        for _ in range(count):
            data = bytearray([randint(0, 255) for _ in range(length)])
            sink.sendto(data, (address, port))

    async def warmup(self):
        """
        If server ports were received, send reverse direction dummy
        packets to force NAT traversal
        """
        await super().warmup()

        if self.server_rtp is not None:
            self.logger.info('sending warmup RTP uplink traffic')
            self.send_upstream(self.rtp_sink, self.connection.host, self.server_rtp)

        if self.server_rtcp is not None:
            self.logger.info('sending warmup RTCP uplink traffic')
            self.send_upstream(self.rtcp_sink, self.connection.host, self.server_rtcp)

    async def send_rtcp_report(self, rtcp: RTCP):
        if self.rtcp_sender:
            self.rtcp_sink.sendto(bytes(rtcp), self.rtcp_sender)
